#pragma once

#include <vector>
#include <SFML/Graphics.hpp>
#include <nlohmann/json.hpp>
#include <spdlog/spdlog.h>


#define TYPEDEF_VECTOR(NAME, T, N, s) typedef NAME<T> vec##N##s;

#define TYPEDEF_VECTORS(NAME, N) \
	typedef NAME<float> vec##N##; \
	TYPEDEF_VECTOR(NAME, float, N, f) \
	TYPEDEF_VECTOR(NAME, double, N, d) \
	TYPEDEF_VECTOR(NAME, long double, N, ld) \
	TYPEDEF_VECTOR(NAME, size_t, N, sz) \
	TYPEDEF_VECTOR(NAME, int, N, i) \
	TYPEDEF_VECTOR(NAME, unsigned int, N, u) \
	TYPEDEF_VECTOR(NAME, short, N, s) \
	TYPEDEF_VECTOR(NAME, unsigned short, N, us) \
	TYPEDEF_VECTOR(NAME, long, N, l) \
	TYPEDEF_VECTOR(NAME, unsigned long, N, ul) \
	TYPEDEF_VECTOR(NAME, long long, N, ll) \
	TYPEDEF_VECTOR(NAME, unsigned long long, N, ull)




namespace UI
{
	using json = nlohmann::json;

	template<typename T>
	union Vector2 {
		struct { T x, y; };
		struct { T a, b; };
	};

	TYPEDEF_VECTORS(Vector2, 2)

	template<typename T>
	union Vector3 {
		struct { T x, y, z; };
		struct { T r, g, b; };
	};
	TYPEDEF_VECTORS(Vector3, 3)

	template<typename T>
	union Vector4 {
		struct { T x, y, z, w; };
		struct { T r, g, b, a; };
		struct { T left, top, right, bottom; };
	};
	TYPEDEF_VECTORS(Vector4, 4)



	class RenderCall
	{
		std::unique_ptr<sf::Drawable> drawable;
		sf::RenderStates states;

	public:
		explicit RenderCall(const sf::RenderStates& states, sf::Drawable* ptr) : drawable(ptr), states(states)
		{
			if (!drawable) { throw std::runtime_error("RenderCall::RenderCall(): Drawable is null"); }
		}

		void draw(sf::RenderTarget& target, const sf::RenderStates& states) const { target.draw(*drawable, states); }
	};

	enum class LayoutSizes : uint8_t
	{
		None = 0,
		Min,
		Max,
		Fixed,
	};

	enum class LayoutAlignment : uint8_t
	{
		None = 0,
		Left,
		Right,
		Top,
		Bottom,
		Center,
	};

	enum class LayoutDirection : uint8_t
	{
		None = 0,
		LeftToRight,
		RightToLeft,
		TopToBottom,
		BottomToTop,
	};

	enum class LayoutPosition : uint8_t
	{
		None = 0,
		Absolute,
		Left,
		Right,
		Top,
		Bottom,
		Center,
	};

	struct Layout
	{
		struct {
			LayoutSizes rule = LayoutSizes::None;
			vec2f minSize = {0.f, 0.f};
			vec2f maxSize = {0.f, 0.f};
		} size;

		struct {
			LayoutAlignment rule = LayoutAlignment::None;
			vec4f padding = {0.f, 0.f, 0.f, 0.f};
			vec4f margin = {0.f, 0.f, 0.f, 0.f};
		} alignment;

		struct {
			LayoutDirection rule = LayoutDirection::None;
			vec2f spacing = {0.f, 0.f};
		} direction;

		struct {
			LayoutPosition rule = LayoutPosition::None;
			vec2f position = {0.f, 0.f};
		} position;
	};

	class View
	{
	protected:
		std::vector<RenderCall> m_RenderCalls;
		Layout m_Layout;

	public:
		virtual ~View() = default;

		[[nodiscard]] const std::vector<RenderCall>& renderCalls() const { return m_RenderCalls; }

		virtual bool update() { return false; }

		void draw()
		{
			if (!m_RenderCalls.empty()) { m_RenderCalls.clear(); }
			body();
		}

		virtual void body() = 0;
	};

	class Rectangle : public View
	{

		vec2f m_Size;
		sf::Color m_Color;
		sf::Color m_BorderColor;
		float m_BorderThickness = 0;


		void setBorderColor(const sf::Color& color) { m_BorderColor = color; }
		void setBorderThickness(float thickness) { m_BorderThickness = thickness; }
		void setColor(const sf::Color& color) { m_Color = color; }
		void setSize(const vec2f& size) { m_Size = size; }

		void body() override
		{
			auto m_Shape = new sf::RectangleShape(sf::Vector2f{m_Size.x, m_Size.y});
			m_Shape->setFillColor(m_Color);
			if (m_BorderThickness > 0)
			{
				m_Shape->setOutlineThickness(m_BorderThickness);
				m_Shape->setOutlineColor(m_BorderColor);
			}
			m_RenderCalls.emplace_back(sf::RenderStates(), m_Shape);
		}

		Rectangle(const vec2f& size) : m_Size(size) {}
	};

	class Renderer
	{
	public:
		explicit Renderer(View* _view) : view(_view) { view->body(); }

		std::unique_ptr<View> view;

		void update() const
		{
			if (view->update())
				view->draw();
		}

		void render(sf::RenderTarget& target, const sf::RenderStates& states) const
		{
			for (const auto& renderCall : view->renderCalls()) { renderCall.draw(target, states); }
		}
	};
}
